#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 3A.mr - Reservoir Sampling distribué - énoncé - correction
# 
# Correction.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## création d'un fichier à sampler

# In[2]:


with open("sample4.txt", "w", encoding="utf8") as f:
    for i in range(0,100000):
        f.write("{0}\t{1}{0}\n".format(i, chr(i%26 + 65)))
    f.write("100001\tAAAAAA")


# In[3]:


get_ipython().run_line_magic('load_ext', 'pyensae')
get_ipython().run_line_magic('head', 'sample4.txt')


# ## connexion

# In[4]:


import os
blobhp = {}
if "HDCREDENTIALS" in os.environ:
    blobhp["blob_storage"], blobhp["password1"], blobhp["hadoop_server"], blobhp["password2"], blobhp["username"] = \
        os.environ["HDCREDENTIALS"].split("**")
    r = type(blobhp)
else:
    from pyquickhelper.ipythonhelper import open_html_form
    params={"blob_storage":"", "password1":"", "hadoop_server":"", "password2":"", "username":"axavier"}
    r = open_html_form(params=params,title="server + hadoop + credentials", key_save="blobhp")
r


# In[5]:


import pyensae
get_ipython().run_line_magic('load_ext', 'pyensae')
get_ipython().run_line_magic('load_ext', 'pyenbc')
blobstorage = blobhp["blob_storage"]
blobpassword = blobhp["password1"]
hadoop_server = blobhp["hadoop_server"]
hadoop_password = blobhp["password2"]
username = blobhp["username"]


# In[6]:


client, bs = get_ipython().run_line_magic('hd_open', '')
client, bs


# ## upload du fichier

# In[7]:


get_ipython().run_line_magic('blob_up', 'sample3.txt /$PSEUDO/sampling/sample4.txt')


# In[8]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/sampling')


# ## Code python pour le reservoir sampling

# In[9]:


ensemble = [ "%d%s" % (i, chr(i%26 + 97)) for i in range(0,10000)]
ensemble[:5]


# In[10]:


import random
def reservoir_sampling(ensemble, k):
    N = len(ensemble)
    echantillon = []
    for i, e in enumerate(ensemble):
        if len(echantillon) < k:
            echantillon.append(e)
        else:
            j = random.randint(0, i)
            if j < k:
                echantillon[j] = e
    return echantillon

reservoir_sampling(ensemble, 10)


# ## python à jython
# 
# On s'assure que le code précédent fonctionne en jython (python compilé en java). On s'inspire pour cela de la documentation [jython-udfs](https://pig.apache.org/docs/r0.12.0/udf.html#jython-udfs).

# ### On créé d'abord un script PIG pour récupérer le schema et les premières lignes

# In[11]:


get_ipython().run_cell_magic('PIG', 'sample_explore.pig', "\nensemble = LOAD '$CONTAINER/$PSEUDO/sampling/sample4.txt' \n        USING PigStorage('\\t') AS (x:int, v:chararray) ;\nDESCRIBE ensemble;\nens_group = GROUP ensemble ALL;\nDESCRIBE ens_group;\nsampled = FOREACH ens_group GENERATE FLATTEN(ensemble);\nDESCRIBE sampled;\n\n--ens10 = LIMIT ensemble 10;\n--ens_group10 = LIMIT en_group10 ;\n--DUMP ens10;\n--DUMP ens_group10;\n")


# Si la fonction suivante provoque une erreur ::
# 
#     AzureException: STATUS: 403, JSON: Expecting value: line 1 column 1 (char 0)
#     <Response [403]>
#     unable to submit job: sample_explore.pig
#     
# Vérifier les identifiants utilisés pour se connecter.

# In[12]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'sample_explore.pig')
jid


# In[13]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
(st["id"],st["percentComplete"],st["completed"],
st["status"]["jobComplete"],st["status"]["state"])


# La sortie standard contient les informations souhaitées :

# In[14]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 5')


# Et la sortie du second dump :: 
#     
#     (all,{(100001,AAAAAA),(99999,D99999),(99998,C99998)...

# ### Le code Jython

# In[15]:


import pyensae


# In[16]:


get_ipython().run_cell_magic('PYTHON', 'reservoir_sampling.py', '\nimport random\n\n@outputSchemaFunction("rsSchema")\ndef reservoir_sampling(ensemble):\n    ensemble = eval(ensemble)\n    k = 10\n    N = len(ensemble)\n    echantillon = []\n    for i, e in enumerate(ensemble):\n        if len(echantillon) < k:\n            echantillon.append(e)\n        else:\n            j = random.randint(0, i)\n            if j < k:\n                echantillon[j] = e\n    return echantillon\n\n@schemaFunction("rsSchema")\ndef rsSchema(input):\n    return input\n')


# In[17]:


get_ipython().run_cell_magic('jython', 'reservoir_sampling.py reservoir_sampling', '{(100001,"AAAAAA"),(99999,"D99999"),(99998,"C99998")}\n')


# On ajoute le code jython au script précédent :

# In[18]:


get_ipython().run_cell_magic('PIG', 'sample_explore_complete.pig', "\nREGISTER '$CONTAINER/$SCRIPTPIG/reservoir_sampling.py' using jython as myrs;\n\nensemble = LOAD '$CONTAINER/$PSEUDO/sampling/sample4.txt' \n        USING PigStorage('\\t') AS (x:int, v:chararray) ;\nDESCRIBE ensemble;\nens_group = GROUP ensemble ALL;\nDESCRIBE ens_group;\nsampled = FOREACH ens_group GENERATE FLATTEN(myrs.reservoir_sample(ensemble));\nDESCRIBE sampled;\n\nSTORE sampled INTO \nINTO '$CONTAINER/$PSEUDO/sampling/sample_rs.txt' USING PigStorage();\n")


# In[19]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'sample_explore_complete.pig -d reservoir_sampling.py')
jid


# In[20]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
(st["id"],st["percentComplete"],st["completed"],
st["status"]["jobComplete"],st["status"]["state"])


# In[21]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 100')


# A corriger plus tard. Dans l'immédiat, on utilisera la librairie [datafu](https://datafu.incubator.apache.org/docs/datafu/guide/sampling.html). Si le cluster ne reconnaît pas la librairie, voir la section java pour comprendre comment l'importer. On la déclare dans le script par l'instruction ``REGISTER``.

# In[22]:


get_ipython().run_cell_magic('PIG', 'sample_explore_datafu.pig', "\nREGISTER '$CONTAINER/$PSEUDO/sampling/datafu-1.2.0.jar';\nDEFINE RS datafu.pig.sampling.ReservoirSample('1000');\n\nensemble = LOAD '$CONTAINER/$PSEUDO/sampling/sample4.txt' \n        USING PigStorage('\\t') AS (x:int, v:chararray) ;\nDESCRIBE ensemble;\nens_group = GROUP ensemble ALL;\nDESCRIBE ens_group;\nsampled = FOREACH ens_group GENERATE FLATTEN(RS(ensemble));\nDESCRIBE sampled;\n\nSTORE sampled \nINTO '$CONTAINER/$PSEUDO/sampling/sample_datafu_rs.txt' USING PigStorage();\n")


# In[23]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'sample_explore_datafu.pig')
jid


# In[24]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
(st["id"],st["percentComplete"],st["completed"],
st["status"]["jobComplete"],st["status"]["state"])


# In[25]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 100')


# In[26]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/sampling/sample_datafu')


# ## version distribuée
# 
# Astuce : on distribue puis on recombine les échantillons en faisant un dernier reservoir sampling mais pondéré. Comment distribuer ? Le second sampling est remplacé par une méthode d'échantillonage classique car le reservoir sampling pondéré n'est pas disponible dans la librairie datafu version 1.2.0.

# In[27]:


get_ipython().run_cell_magic('PIG', 'sample_explore_datafu_dist.pig', "\nREGISTER '$CONTAINER/$PSEUDO/sampling/datafu-1.2.0.jar';\nDEFINE RS datafu.pig.sampling.ReservoirSample('1000');\nDEFINE WeightedSample datafu.pig.sampling.WeightedSample();\n\nensemble = LOAD '$CONTAINER/$PSEUDO/sampling/sample4.txt' \n        USING PigStorage('\\t') AS (x:int, v:chararray) ;\nDESCRIBE ensemble;\nkeys = FOREACH ensemble GENERATE x, v, x%10 AS key;\nDESCRIBE keys;\nens_group = GROUP keys BY key ;\nDESCRIBE ens_group;\nsampled = FOREACH ens_group GENERATE COUNT(keys) AS weigth, FLATTEN(RS(keys));\nDESCRIBE sampled;\nwsampled = FOREACH (GROUP sampled ALL) GENERATE FLATTEN(WeightedSample(sampled, 0, 1000));\nDESCRIBE wsampled;\n\nSTORE wsampled \nINTO '$CONTAINER/$PSEUDO/sampling/sample_datafu_rs_dist2.txt' USING PigStorage();\n")


# In[28]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'sample_explore_datafu_dist.pig')
jid


# In[29]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
(st["id"],st["percentComplete"],st["completed"],
st["status"]["jobComplete"],st["status"]["state"])


# In[30]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 10')


# In[31]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/sampling/sample_datafu_rs_dist2')


# In[32]:


df = get_ipython().run_line_magic('blob_head', '/$PSEUDO/sampling/sample_datafu_rs_dist2.txt -m')


# In[33]:


df.head()


# ## version distribuée améliorée
# 
# Le problème de la version précédente : chaque sous-ensemble traité d'un seul bloc utilise une séquence de nombres aléatoires sur laquelle on ne connaît pas grand chose. Si les mêmes *seed* sont utilisées, il est possible que les séquences, même si elles simulent le hasard, soient extrêmement corrélées entre chaque bloc. Il faut remédier à cela.
# 
# Il faut également s'assurer que chaque bloc n'est pas *skewed*.

# In[34]:


get_ipython().run_cell_magic('PIG_azure', 'script_rs.pig', "\nREGISTER '$CONTAINER/$PSEUDO/sampling/datafu-1.2.0.jar';\nDEFINE MD5 datafu.pig.hash.MD5();\nDEFINE RS datafu.pig.sampling.ReservoirSample('1000');\nDEFINE WeightedSample datafu.pig.sampling.WeightedSample();\n\nensemble = LOAD '$CONTAINER/$PSEUDO/sampling/sample4.txt' \n        USING PigStorage('\\t') AS (x:int, v:chararray) ;\nDESCRIBE ensemble;\n\nens_group = GROUP ensemble BY (x,v);\nDESCRIBE ens_group;\n\ncompte_group = FOREACH ens_group \n            GENERATE group.x AS x, \n                     group.v AS v, \n                     COUNT(ensemble) AS nb_ligne ;\nDESCRIBE compte_group;\n\nhash_group = FOREACH compte_group \n                GENERATE x, v, nb_ligne,\n                        SUBSTRING(MD5(v), 0, 1) AS val;\nDESCRIBE hash_group;  \n\ngroup_hash = GROUP hash_group BY val ;\nDESCRIBE group_hash;\n\nrs_parall = FOREACH group_hash GENERATE\n                    COUNT(hash_group) AS nb_hash,\n                    FLATTEN(RS(hash_group)) ;\nDESCRIBE rs_parall;\n\nwsampled = FOREACH (GROUP rs_parall ALL) GENERATE FLATTEN(WeightedSample(rs_parall, 0, 1000));\nDESCRIBE wsampled;\n\nSTORE wsampled \nINTO '$CONTAINER/$PSEUDO/sampling/sample_distributed_hash.txt' USING PigStorage();\n")


# In[35]:


jid=get_ipython().run_line_magic('hd_pig_submit', 'script_rs.pig')
jid


# In[36]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
(st["id"],st["percentComplete"],st["completed"],
st["status"]["jobComplete"],st["status"]["state"])


# In[37]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 10')


# In[38]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/sampling/sample_distributed_hash.txt')


# In[39]:


df =get_ipython().run_line_magic('blob_head', '/$PSEUDO/sampling/sample_distributed_hash.txt -m')
df.head()


# In[40]:


get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/sampling/sample_distributed_hash.txt sample_distributed_hash.txt')


# In[41]:


get_ipython().run_line_magic('head', 'sample_distributed_hash.txt')


# In[42]:





# ## version java
# 
# On s'inspire de l'exemple suivant [Sampling](http://datafu.incubator.apache.org/docs/datafu/guide/sampling.html).
#  On télécharge [datafu 1.2](http://datafu.incubator.apache.org/docs/datafu/) depuis [Maven](http://mvnrepository.com/artifact/com.linkedin.datafu/datafu/1.2.0). Ce n'est pas la dernière version mais suivre les  instructions pour *builder* datafu (voir [documentation](http://datafu.incubator.apache.org/docs/datafu/1.2.0/)). En particulier, la version pondérée du reservoir sampling n'est pas disponible (voir [history](https://github.com/apache/incubator-datafu/commits/master/datafu-pig/src/main/java/datafu/pig/sampling/WeightedReservoirSample.java), la version 1.2.0 est sorti en décembre 2013).
#  
# L'implémentation [java](https://github.com/apache/incubator-datafu/blob/master/datafu-pig/src/main/java/datafu/pig/sampling/ReservoirSample.java) n'a pas l'air de résoudre un problème qui peut survenir si la taille de l'échantillon demandée est trop grande. Voir section suivante.

# In[43]:


import pyensae.datasource
pyensae.datasource.download_data("datafu-1.2.0.jar", url="http://central.maven.org/maven2/com/linkedin/datafu/datafu/1.2.0/")


# In[44]:


get_ipython().run_line_magic('blob_up', 'datafu-1.2.0.jar /$PSEUDO/sampling/datafu-1.2.0.jar')


# In[45]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/sampling')


# In[46]:


get_ipython().run_cell_magic('PIG_azure', 'sample.pig', "\nREGISTER '$CONTAINER/$PSEUDO/sampling/datafu-1.2.0.jar';\n\nDEFINE RS datafu.pig.sampling.ReservoirSample('1000');\n\ndset = LOAD '$CONTAINER/$PSEUDO/sampling/sample4.txt' \n        USING PigStorage('\\t') AS (x:int, v:chararray) ;\nsampled = FOREACH (GROUP dset ALL) GENERATE FLATTEN(RS(dset));\nSTORE sampled INTO '$CONTAINER/$PSEUDO/sampling/out_sampled_rs4_2015.txt' USING PigStorage() ;\n")


# In[47]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'sample.pig')


# In[48]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
st["id"],st["percentComplete"],st["completed"],st["status"]["jobComplete"],st["status"]["state"]


# In[49]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 10')


# In[50]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/sampling')


# In[51]:


get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/sampling/out_sampled_rs4_2015.txt out_sampled_rs4_2015.txt -o')


# In[52]:


get_ipython().run_line_magic('head', 'out_sampled_rs4_2015.txt')


# ## fin

# In[53]:


get_ipython().run_line_magic('blob_close', '')


# ## version avec itérateur

# In[54]: